# ------------------------------------------------------------------------------
#
# FishFa40.py                                            Ulrich Mueller 03.01.10
#                                                        PCS1700    Python 3.1.1
# Modul FishFa40 mit Funktionen zum Betrieb der fischertechnik Interfaces
# Klasse FishFace : Kapselung und Erweiterung der Funktionen von umFish40.DLL
#                   benoetigt umFish40.DLL
# Klasse FishFaceException : Raise FishFace Ausnahmen
# Klasse FishRobot :         Betrieb eines Robots mit RobMotoren
# Klasse FishStep  :         Betrieb von Schrittmotoren, einzeln oder im XY-Verbund
#
# ---------------------------------------------------------------------------------------

from ctypes import *

class FishFaceException(IOError):
  'Allgemeiner FishFace Fehler'

class FishFace(object):

  iHandle      = -1          
                             # gleichzeitig Open-Merker (no Open = -1)
  PollInterval = 12
  
# ---- symbolische Konstanten -------------------
  Fehler       = 0x1FFFF
  Ein          = 1
  Aus          = 0
  Links        = 1
  Rechts       = 2

  Ende         = 0
  Time         = 1
  ESC          = 3

  ftROBO_first_USB       =   0
  ftIntelligent_IF       =  10
  ftIntelligent_IF_Slave =  20
  ftROBO_IF_IIM          =  50
  ftROBO_IF_USB          =  60
  ftROBO_IF_COM          =  70
  ftROBO_IF_Over_RF      =  80
  ftROBO_Extension       =  90
  ftROBO_RF_Datalink     = 110
  
# --- interne Werte ---------------------------
  ESCAPE       = 27
  
  def __init__(self):
    self.iHandle = -1
    
# --- oeffentliche Methoden und Eigenschaften ---

  def OpenInterface(self, TypNr=0, Nr=0):
    "Verbindung zum Interface herstellen"
    if TypNr == 10 or TypNr == 20 or TypNr == 70:
      self.iHandle = windll.umfish40.rbOpenInterfaceCOM(TypNr, Nr, 0)
      print ("COM", TypNr, Nr)
    else:
      self.iHandle = windll.umFish40.rbOpenInterfaceUSB(TypNr, Nr)
      print ("USB", TypNr, Nr)

    if self.iHandle == self.Fehler:
      raise (FishFaceException, 'InterfaceProblem.Open')  
    self.SetOutputs(0)        

  def ClearCounter(self, InputNr):
    "Loeschen eines Counters"
    windll.umFish40.rbSetCounter(self.iHandle, InputNr, 0)

  def ClearCounters(self):
    "Loeschen aller Counter"
    windll.umFish40.rbClearCounters(self.iHandle)

  def ClearMotors(self):
    "Loeschen aller M-Ausgaenge"
    windll.umFish40.rbSetMotors(self.iHandle, 0)
    
  def CloseInterface(self):
    "Schliessen Interface"
    windll.umFish40.rbCloseInterface(self.iHandle)
    self.iHandle = -1

  def Finish(self, InputNr=0):
    "Feststellen Endewunsch (Escape, InputNr)"
    if windll.user32.GetAsyncKeyState(self.ESCAPE) != 0: return 1
    if InputNr > 0 and self.GetInput(InputNr): return 1
    return 0
    
  def GetAnalog(self, AnalogNr):
    "Auslesen Analogwert AX, AY, Extensions (1 - 2 (5))"
    return windll.umFish40.rbGetAnalog(self.iHandle, AnalogNr)
  
  def GetCounter(self, InputNr):
    "Auslesen eines Counters"
    return windll.umFish40.rbGetCounter(self.iHandle, InputNr)

  def GetInput(self, InputNr):
    "Bestimmen Status I-Eingang"
    return windll.umFish40.rbGetInput(self.iHandle, InputNr)
  
  def GetInputs(self):
    "Auslesen Status aller I-Eingaenge"
    return windll.umFish40.rbGetInputs(self.iHandle)

  def GetOutputs(self):
    "Auslesen Status aller M-Ausgaenge"
    return windll.umFish40.rbGetMotors(self.iHandle)

  def GetVoltage(self, VoltNr):
    "Auslesen Voltwert A1, A2, AV, AZ (1 - 4)"
    return windll.umFish40.rbGetVoltage(self.iHandle, VoltNr)
  
  def Pause(self, mSek):
    "Programm anhalten"    
    sZeit = windll.kernel32.GetTickCount()
    while (windll.kernel32.GetTickCount() - sZeit) < mSek:
      if windll.user32.GetAsyncKeyState(self.ESCAPE) != 0 : return
      windll.kernel32.Sleep(self.PollInterval)

  def SetCounter(self, InputNr, Wert):
    "Setzen Counter"
    windll.umFish40.rbSetCounter(self.iHandle, InputNr, Wert)
    
  def SetLamp(self, LampNr, OnOff):
    "Setzen eines O-Ausganges"
    windll.umFish40.rbSetLamp(self.iHandle, LampNr, OnOff)
    
  def SetOutputs(self, Wert):
    "Setzen aller M-Ausgaenge"
    windll.umFish40.rbSetMotors(self.iHandle, Wert)

  def SetMotor(self, MotorNr, Direction, Speed=131071, Counter=131071):
    "Setzen eines M-Ausganges"      
    if Speed == self.Fehler:
      windll.umFish40.rbSetMotor(self.iHandle, MotorNr, Direction)
      return
      
    if Counter == self.Fehler:
      windll.umFish40.rbSetMotorEx(self.iHandle, MotorNr, Direction, Speed)
      return

    windll.umFish40.rbRobMotor(self.iHandle, MotorNr, Direction, Speed, Counter) 

  def SetMotors(self, MotorStatus, SpeedStatus=131071, ModeStatus=131071):
    "Setzen aller M-Ausgaenge"
    if SpeedStatus == self.Fehler:
      windll.umFish40.rbSetMotors(self.iHandle, MotorStatus)
      return
    if ModeStatus == self.Fehler:
      windll.umFish40.rbSetMotorsEx(self.iHandle, MotorStatus, SpeedStatus)
      return
    windll.umFish40.rbRobMotors(self.iHandle, MotorStatus, SpeedStatus, ModeStatus)
  
  def Version(self):
    "Version der Klasse"
    return "4.1.01"

  def WaitForChange(self, InputNr, NrOfChanges, TermInputNr = 0):
    "Warten auf NrOfChanges Impulse von InputNr"
    windll.umFish40.rbSetCounter(self.iHandle, InputNr, 0)
    while windll.umFish40.rbGetCounter(self.iHandle, InputNr) < NrOfChanges:
      if windll.user32.GetAsyncKeyState(self.ESCAPE) != 0 : return
      if TermInputNr>0:
        if self.GetInput(TermInputNr): return
        windll.kernel32.Sleep(self.PollInterval)
   
  def WaitForHigh(self, InputNr):
    "Warten auf einen false/true-Durchgang von InputNr"
    self.WaitForInput(InputNr, 0)
    self.WaitForInput(InputNr, 1)

  def WaitForInput(self, InputNr, OnOff=1):
    "Warten auf InputNr = OnOff"
    while not (self.GetInput(InputNr) == OnOff):
      windll.kernel32.Sleep(self.PollInterval)
      if windll.user32.GetAsyncKeyState(self.ESCAPE) != 0 : return

  def WaitForLow(self, InputNr):
    "Warten auf einen true/false-Durchgang von InputNr"
    self.WaitForInput(InputNr, 1)
    self.WaitForInput(InputNr, 0)

  def WaitForMotors(self, mSek, MotorNrs):
    "Warten auf MotorReady-Ereignis oder Ablauf mSek"
    t = windll.kernel32.GetTickCount() + mSek
    while mSek == 0 or t>windll.kernel32.GetTickCount():
      c = 0
      for i in MotorNrs : c += windll.umFish40.rbGetModeStatus(self.iHandle, i)
      if c == 0 : return self.Ende
      if windll.user32.GetAsyncKeyState(self.ESCAPE) != 0 :
        if i in MotorNrs : windll.umFish40.rbSetMotor(self.iHandle, i, 0)
        return self.ESC
        windll.kernel32.Sleep(self.PollInterval)
      return self.Time

  def WaitForPositionDown(self, InputNr, ActPosition, TargetPosition, TermInputNr=0):
    "Warten auf TargetPosition = ActPosition - InputNr-Impulse"
    self.WaitForChange(InputNr, ActPosition-TargetPosition, TermInputNr)
    return ActPosition-self.GetCounter(InputNr)

  def WaitForPositionUp(self, InputNr, ActPosition, TargetPosition, TermInputNr=0):
    "Warten auf TargetPosition = ActPosition + InputNr-Impulse"
    self.WaitForChange(InputNr, TargetPosition-ActPosition, TermInputNr)
    return ActPosition+self.GetCounter(InputNr)

  def WaitForTime(self, mSek):
    "Programm anhalten (wie Pause)"
    self.Pause(mSek)


# ------------------------------------------------------------------------------
# Klasse FishRobot
# ------------------------------------------------------------------------------

class FishRobot(FishFace):
  "Betrieb eines Industry Robot aehnlichen Robots mit 'RobMotoren' "

# --- MotCntl : Liste der eingesetzten Motoren - Nr, maxPos, actPos ---
  MotCntl = []

  def __init__(self, MotList):
    FishFace.__init__(self)
    for m in MotList:
      self.MotCntl.append([m[0],m[1],0])

  def PosPrint(self, PosList):
    "Ausgabe der aktuellen Robot-Position"
    print ("PosPrint : ", PosList)

  def MoveHome(self):
    "Anfahren der Home-Position"
    Mots = []
    for m in self.MotCntl: Mots.append(m[0])
    for nr in Mots: self.SetMotor(nr, self.Links, 15, 9999)
    self.WaitForMotors(0, Mots)

  def MoveTo(self, PosHook, PosList):
    "Simultanes Anfahren einer vorgegebenen Position"
  # --- Starten der Motoren -------------------------
    anzPos = len(PosList)
    for i in range(0, anzPos):
      ZielPos = PosList[i]
      if ZielPos > self.MotCntl[i][1]: ZielPos = self.MotCntl[i][1]
      if ZielPos > self.MotCntl[i][2]:
        self.SetMotor(self.MotCntl[i][0], self.Rechts, 15, \
                      ZielPos -self.MotCntl[i][2])
      else:
        ct = self.MotCntl[i][2]-ZielPos
        if ct > 0: self.SetMotor(self.MotCntl[i][0], self.Links, 15, ct)
  # --- Warten auf MotorReady -------------------------
    Mots = []
    for m in self.MotCntl: Mots.append(m[0])
    res = self.Time
    while res == self.Time:
      res = self.WaitForMotors(8*self.PollInterval, Mots)
    # --- Ausgabe der aktuellen Position ---------  
      if PosHook is not None:
        ActPos = []
        for i in range(0, anzPos):
          iNr = (Mots[i]-1)*2+2
          ZielPos = PosList[i]
          if ZielPos > self.MotCntl[i][2]:
            ActPos.append(ZielPos - self.GetCounter(iNr))
          else:
            ActPos.append(ZielPos + self.GetCounter(iNr))                         
        PosHook(ActPos)
  # --- Aktualisieren MotCntl ----------------------
    for i in range(0, anzPos):
      iNr = (Mots[i]-1)*2+2
      ZielPos = PosList[i]
      if res == self.Ende:
        if ZielPos > self.MotCntl[i][2]:
          self.MotCntl[i][2] = ZielPos + self.GetCounter(iNr)
        else:
          self.MotCntl[i][2] = ZielPos - self.GetCounter(iNr)
      else:
        if ZielPos > self.MotCntl[i][2]:
          self.MotCntl[i][2] = ZielPos - self.GetCounter(iNr)
        else:
          self.MotCntl[i][2] = ZielPos + self.GetCounter(iNr)
  # --- Abschliessende Positionsangabe ------------------------
    if PosHook is not None:
      ActPos = []
      for i in range(0, anzPos): ActPos.append(self.MotCntl[i][2])
      PosHook(ActPos)

# ------------------------------------------------------------------------------------
# Klasse FishStep
# ------------------------------------------------------------------------------------

class FishStep(FishFace):
  "Betrieb von Schrittmotoren : Einzeln und im XY-Verbund"
  # --- Einzelmotor an zwei nebeneinanderliegenden M-Ausgängen
  #     XY-Verbund an drei nebeneinanderliegenden M-Augängen
  # --- feste Geschwindigkeit : PollInterval, Synchron-Betrieb
  #     Masseinheit : Zyklus ab 0
  #     feste Endtaster Zuordnung : M1 = I1, M2 = I3, M3 = I5 ... M8 = I15 ...
  #     MotCntl[[max,act,out]] pro Motor
  
  MotCntl = [[999,0,0],[999,0,0],[999,0,0],[999,0,0],[999,0,0],[999,0,0],[999,0,0],[999,0,0]]
  MotLLL,MotLLR,MotLRL,MotLRR,MotRLL,MotRLR,MotRRL,MotRRR = 0,1,2,3,4,5,6,7
  
  def __init__(self, MotList):
    FishFace.__init__(self)
    for i in range(0,len(MotList)):
      self.MotCntl[MotList[i][0]-1][0] = MotList[i][1]

  EndNrX = lambda self,MotNr: (MotNr-1)*2+1
  EndNrY = lambda self,MotNr: (MotNr+1)*2+1
  Xmax   = lambda self,MotNr: self.MotCntl[MotNr-1][0]
  Ymax   = lambda self,MotNr: self.MotCntl[MotNr+1][0]
  Xact   = lambda self,MotNr: self.MotCntl[MotNr-1][1]
  Yact   = lambda self,MotNr: self.MotCntl[MotNr+1][1]
  Xout   = lambda self,MotNr: self.MotCntl[MotNr-1][2]
  Yout   = lambda self,MotNr: self.MotCntl[MotNr+1][2]
  def XactInc(self,MotNr): self.MotCntl[MotNr-1][1] += 1
  def XactDec(self,MotNr): self.MotCntl[MotNr-1][1] -= 1
  def YactInc(self,MotNr): self.MotCntl[MotNr+1][1] += 1
  def YactDec(self,MotNr): self.MotCntl[MotNr+1][1] -= 1

# -
# ---------- Einzelmotor ----------------------------------------------------------
# -

  def RechtsDreh(self, MotNr):
    "Fahren eines RechtsZyklus"
    self.SetMotor(MotNr, self.Links)
    self.Pause(self.PollInterval)
    self.SetMotor(MotNr+1, self.Links)
    self.Pause(self.PollInterval)
    self.SetMotor(MotNr, self.Rechts)
    self.Pause(self.PollInterval)
    self.SetMotor(MotNr+1, self.Rechts)
    self.Pause(self.PollInterval)
  
  def LinksDreh(self, MotNr):
    "Fahren eines LinksZyklus"
    self.SetMotor(MotNr, self.Rechts)
    self.Pause(self.PollInterval)
    self.SetMotor(MotNr+1, self.Links)
    self.Pause(self.PollInterval)
    self.SetMotor(MotNr, self.Links)
    self.Pause(self.PollInterval)
    self.SetMotor(MotNr+1, self.Rechts)
    self.Pause(self.PollInterval)
    
  def PosPrintX(self, PosX):
    "Ausgabe der aktuellen X-Position"
    print ("PosPrintX : ", PosX)

  def StepHome(self, MotNr):
    "Fahren Motor auf Home-Position"
    while not self.GetInput(self.EndNrX(MotNr)):
      if windll.user32.GetAsyncKeyState(self.ESCAPE) != 0 : return
      self.LinksDreh(MotNr)
    self.MotCntl[MotNr-1][1] = 0        # --- Xact
    self.MotCntl[MotNr-1][2] = 0        # --- Xout

  def StepTo(self, PosHook, MotNr, Xabs):
    "Fahren Motor auf ZielPosition"
    PosCount = 0
    if Xabs > self.Xact(MotNr):
      while self.Xact(MotNr) < Xabs:
        self.RechtsDreh(MotNr)
        self.XactInc(MotNr)
        if Xabs >= self.MotCntl[MotNr-1][0]: return
        if windll.user32.GetAsyncKeyState(self.ESCAPE) != 0 : return
        PosCount += 1
        if PosHook is not None and PosCount == 10:
          PosCount = 0
          PosHook(self.Xact(MotNr))
    else:
      while self.Xact(MotNr) > Xabs:
        self.LinksDreh(MotNr)
        self.XactDec(MotNr)
        if self.GetInput(self.EndNrX(MotNr)): return
        if windll.user32.GetAsyncKeyState(self.ESCAPE) != 0 : return
        PosCount += 1
        if PosHook is not None and PosCount == 10:
          PosCount = 0
          PosHook(self.Xact(MotNr))
    if PosHook is not None:
      PosHook(self.Xact(MotNr))

  def StepDelta(self, PosHook, MotNr, Xrel):
    "Fahren Motor um Xrel-Zyklen"
    self.StepTo(PosHook, MotNr, self.Xact(MotNr)+Xrel)

# ------------- XY-Verbund --------------------------------------------
# --- Zyklen in die acht Richtungen -----------------------------------

  def BereichsLimit(self, MotNr):
    "Pruefen, ob actKoordinaten im Plotbereich liegen"
    if self.Xout(MotNr) or self.Yout(MotNr):
      self.MotCntl[MotNr-1][2] = self.Xact(MotNr) < 0 or self.Xact(MotNr) > self.Xmax(MotNr)
      self.MotCntl[MotNr+1][2] = self.Yact(MotNr) < 0 or self.Yact(MotNr) > self.Ymax(MotNr)
      if self.Xout(MotNr) or self.Yout(MotNr): return
    else:
      self.MotCntl[MotNr-1][2] = self.Xact(MotNr) < 0 or self.Xact(MotNr) > self.Xmax(MotNr)
      self.MotCntl[MotNr+1][2] = self.Yact(MotNr) < 0 or self.Yact(MotNr) > self.Ymax(MotNr)
      
  def Endtaster(self, MotNr):
    "Abfrage Endtaster im laufenden Betrieb"
    if (self.GetInput(self.EndNrX(MotNr)) and self.Xact(MotNr) != 0) or \
         (self.GetInput(self.EndNrY(MotNr)) and self.Yact(MotNr) != 0):
      print ("--- Achtung Dejustierung! ---")
      return 1
    else: return 0

  def PosPrintXY(self, PosX, PosY):
    "Ausgabe der aktuellen XY-Position"
    print ("PosPrintXY : ", PosX, PosY)

  def Zyklus(self, MotNr, *Befehle):
    "Abarbeiten des vorgegebenen Befehlzyklus"
    for b in Befehle:
      O = self.GetOutputs() & [0xFFC0,0xFF03,0xFF0F,0xF03F,0xC0FF][MotNr-1]
      self.SetOutputs(O | [0x15,0x25,0x19,0x29,0x16,0x26,0x1A,0x2A][b] <<
                      [0,2,4,8,16][MotNr-1])
      self.Pause(self.PollInterval)

  def XMinusSchritt(self, MotNr):
    "Fahren in -X Richtung"
    self.XactDec(MotNr)
    self.BereichsLimit(MotNr)
    if self.Xout(MotNr): return
    self.Zyklus(MotNr, self.MotRLR, self.MotLLR, self.MotLRR, self.MotRRR)
    
  def XPlusSchritt(self, MotNr):
    "Fahren in +X Richtung"
    self.XactInc(MotNr)
    self.BereichsLimit(MotNr)
    if self.Xout(MotNr): return
    self.Zyklus(MotNr, self.MotLRR, self.MotLLR, self.MotRLR, self.MotRRR)

  def XpYpDiagonale(self, MotNr):
    "Fahren in +X/+Y Richtung"
    self.XactInc(MotNr)
    self.YactInc(MotNr)
    self.BereichsLimit(MotNr)
    if self.Xout(MotNr):   self.YPlusSchritt(MotNr); return
    elif self.Yout(MotNr): self.XPlusSchritt(MotNr); return    
    self.Zyklus(MotNr, self.MotLRR, self.MotLLL, self.MotRLL, self.MotRRR)
    
  def XpYmDiagonale(self, MotNr):
    "Fahren in +X/-Y Richtung"
    self.XactInc(MotNr)
    self.YactDec(MotNr)
    self.BereichsLimit(MotNr)
    if self.Xout(MotNr):   self.YMinusSchritt(MotNr); return
    elif self.Yout(MotNr): self.XPlusSchritt(MotNr);  return    
    self.Zyklus(MotNr, self.MotRRL, self.MotLRL, self.MotLLR, self.MotRLR, self.MotRRR)
    
  def XmYmDiagonale(self, MotNr):
    "Fahren in -X/-Y Richtung"
    self.XactDec(MotNr)
    self.YactDec(MotNr)
    self.BereichsLimit(MotNr)
    if self.Xout(MotNr):   self.YMinusSchritt(MotNr); return
    elif self.Yout(MotNr): self.XMinusSchritt(MotNr); return    
    self.Zyklus(MotNr, self.MotRLL, self.MotLLL, self.MotLRR, self.MotRRR)

  def XmYpDiagonale(self, MotNr):
    "Fahren in -X/+Y Richtung"
    self.XactDec(MotNr)
    self.YactInc(MotNr) 
    self.BereichsLimit(MotNr)
    if self.Xout(MotNr):   self.YPlusSchritt(MotNr);  return
    elif self.Yout(MotNr): self.XMinusSchritt(MotNr); return    
    self.Zyklus(MotNr, self.MotRLR, self.MotLLR, self.MotLRL, self.MotRRL, self.MotRRR)

  def YMinusSchritt(self, MotNr):
    "Fahren in -Y Richtung"
    self.YactDec(MotNr)
    self.BereichsLimit(MotNr)
    if self.Yout(MotNr): return
    self.Zyklus(MotNr, self.MotRRL, self.MotLRL, self.MotLRR, self.MotRRR)
    
  def YPlusSchritt(self, MotNr):
    "Fahren in +Y Richtung"
    self.YactInc(MotNr)
    self.BereichsLimit(MotNr)
    if self.Yout(MotNr): return
    self.Zyklus(MotNr, self.MotLRR, self.MotLRL, self.MotRRL, self.MotRRR)

# --- Plot-Befehle ----------------------------------------------------------

  def PlotHome(self, MotNr):
    "Fahren Motoren auf Home-Position(native : linke untere Ecke)"
    self.MotCntl[MotNr-1][2] = 0   # --- Xout
    self.MotCntl[MotNr+1][2] = 0   #     Yout
    self.SetOutputs(0)
    
    # --- Beide Motoren gegen Endtaster ---
    while not (self.GetInput(self.EndNrX(MotNr)) or self.GetInput(self.EndNrY(MotNr))):
      if windll.user32.GetAsyncKeyState(self.ESCAPE) != 0 : return
      self.Zyklus(MotNr, self.MotRLL, self.MotLLL, self.MotLRR, self.MotRRR)
    # --- Motor X gegen Endtaster ---------
    while not (self.GetInput(self.EndNrX(MotNr))):
      if windll.user32.GetAsyncKeyState(self.ESCAPE) != 0 : return
      self.Zyklus(MotNr, self.MotRLR, self.MotLLR, self.MotLRR, self.MotRRR)
    # --- Motor X frei von Endtaster -------
    while (self.GetInput(self.EndNrX(MotNr))):
      self.Zyklus(MotNr, self.MotLRR, self.MotLLR, self.MotRLR, self.MotRRR)
    # --- Motor Y gegen Endtaster -------
    while not (self.GetInput(self.EndNrY(MotNr))):
      if windll.user32.GetAsyncKeyState(self.ESCAPE) != 0 : return
      self.Zyklus(MotNr, self.MotRRL, self.MotLRL, self.MotLRR, self.MotRRR)
    # --- Motor Y frei von Endtaster -------
    while (self.GetInput(self.EndNrY(MotNr))):
      self.Zyklus(MotNr, self.MotLRR, self.MotLRL, self.MotRRL, self.MotRRR)
    # --- Motoren 2 Schritt Abstand von Ende -------------------------------
    self.XpYpDiagonale(MotNr)
    self.XpYpDiagonale(MotNr)
    self.MotCntl[MotNr-1][1] = 0   # --- Xact
    self.MotCntl[MotNr+1][1] = 0   #     Yact
  
  def PlotDelta(self, PosHook, MotNr, Xrel, Yrel):
    "Fahren Motoren um Xrel/Yrel-Werte(Plotzyklen)"
    Xdelta   = abs(Xrel)  # --- Anzahl Plotzyklen
    Ydelta   = abs(Yrel)
    Xnr=Ynr  = 0          # --- Zyklen-Zaehler
    Delta    = 0          # --- Abstand zur optimalen Geraden
    PosCount = 0          # --- Zaehler fuer PosAusgabe

    while not (Xnr == Xdelta and Ynr == Ydelta):
      if windll.user32.GetAsyncKeyState(self.ESCAPE) != 0 : return
      if self.Endtaster(MotNr) : return
      if abs(Delta + Ydelta) < abs(Delta + Ydelta - Xdelta):
        if Xrel > 0: self.XPlusSchritt(MotNr)
        else:        self.XMinusSchritt(MotNr)
        Xnr   += 1
        Delta += Ydelta
      elif abs(Delta - Xdelta) < abs(Delta + Ydelta - Xdelta):
        if Yrel > 0: self.YPlusSchritt(MotNr)
        else:        self.YMinusSchritt(MotNr)
        Ynr   += 1
        Delta -= Xdelta
      else:
        if Xrel > 0:
          if Yrel > 0: self.XpYpDiagonale(MotNr)
          else:        self.XpYmDiagonale(MotNr)
        else:
          if Yrel > 0: self.XmYpDiagonale(MotNr)
          else:        self.XmYmDiagonale(MotNr)
        Xnr   += 1
        Ynr   += 1
        Delta += Ydelta - Xdelta
      PosCount += 1
      if PosHook is not None and PosCount == 10:
        PosCount = 0
        PosHook(self.Xact(MotNr), self.Yact(MotNr))
    if PosHook is not None:
      PosHook(self.Xact(MotNr), self.Yact(MotNr))

  def PlotTo(self, PosHook, MotNr, Xabs, Yabs):
    "Fahren Motoren auf Xabs/Yabs Position(Plotzyklen ab linke untere Ecke"
    self.PlotDelta(PosHook, MotNr, Xabs-self.Xact(MotNr), Yabs-self.Yact(MotNr))
